/* oracle_get_series, by William Menke, November, 1992

*/

#define TRUE (1)
#define FALSE (0)

#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <math.h>
#include <rpc/rpc.h>
#include <pwd.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <signal.h>
#include "parse.h"

FILE           *fp, *fpw;

char           *malloc(), *realloc(), the_char;

#define MAX_STRING_LENGTH   	(256)
#define SHORT_STRING_LENGTH   	(40)
#define VIEW_DESCRIPTION_LENGTH (1024)
#define MAX_FIELDS 		(100)
#define MAX_RELATED 		(10)
#define MAX_VARIABLES 		(20)
#define MAX_VIEWS 		(50)
#define MAX_IDS			(5)
#define METHOD_BIN 		(0)
#define METHOD_SHORT_ID 	(1)
#define METHOD_LONG_ID 		(2)
#define METHOD_STRING_ID 	(3)
#define METHOD_RECORD_ID	(4)

#define SHORT_TYPE 		(0)
#define LONG_TYPE 		(1)
#define FLOAT_TYPE 		(2)
#define DOUBLE_TYPE 		(3)
#define STRING_TYPE 		(4)
#define NULL_RECORD 		(-1)
#define LINE_LENGTH 		(4096)
#define NAN_SHORT 		(-32768)
#define NAN_LONG 		(-2147483648)

#define RIGHT_PAREN	")"
#define LEFT_PAREN	"("
#define STR_WILDCARD	"%"
#define STR_OPERATOR	"like"
#define SQL_TERMINATOR	";"
#define FIELD_SEPARATOR '\t'
#define MAX_OPTIONS	(20)

char           *cmd[MAX_OPTIONS];

struct sql_ascii_rec {
    char            name[40];
    long            type;
    long            length;
    long            index;
    long            offset;
    char            value[512];
}               field[256];
long            sql_par_n_var;
struct sql_ascii_rec *sql_par_table;
long            query_is_true;


int             n_views;
char           *buffer;

struct view_rec {
    char            view_name[SHORT_STRING_LENGTH];
    int             method;
    char            description[VIEW_DESCRIPTION_LENGTH];
    char            sql_server[MAX_STRING_LENGTH];
    char            shell[MAX_STRING_LENGTH];
    char            reformatter[MAX_STRING_LENGTH];
    char            file_field[MAX_STRING_LENGTH];
    char            hitmap_path[MAX_STRING_LENGTH];
    char            bin_field[MAX_STRING_LENGTH];
    char            extra_tables[MAX_STRING_LENGTH];
    int             n_fields;
    int             lat_field;
    int             lon_field;
    int             buffer_length;
    int 	    longitude_is_neg;

    struct field_rec {
	char            name[SHORT_STRING_LENGTH];
	int             type;
	int             length;
	int             offset;
	char            database_name[SHORT_STRING_LENGTH];
	char            database_table_name[SHORT_STRING_LENGTH];
	char            database_field_name[SHORT_STRING_LENGTH];
	char            comment[MAX_STRING_LENGTH];
    }               f[MAX_FIELDS];

    char           *sql_join;
    int             n_ids;
    char           *sql_id[MAX_IDS];
    int             n_related;

    struct related_rec {
	char            view_name[SHORT_STRING_LENGTH];
	int             view_index;
	int             n_variables;
	int             id;
	struct table_rec {
	    char            variable_name[SHORT_STRING_LENGTH];
	    char            view_field_name[SHORT_STRING_LENGTH];
	    char            format[SHORT_STRING_LENGTH];
	    int             field_index;
	}               v[MAX_VARIABLES];
    }               r[MAX_RELATED];
}               view[MAX_VIEWS];

char            vlist[MAX_IDS][MAX_VARIABLES][SHORT_STRING_LENGTH];
int             n_vars[MAX_IDS];

short           my_short;
long            my_long;
float           my_float;
double          my_double;

float           nandiv, pinfinity, minfinity;
char		lex_field_lookup_buf[MAX_STRING_LENGTH];
int		lex_view;


main(argc, argv)
    int             argc;
    char           *argv[];
{

	char	error_message[MAX_STRING_LENGTH];
	long	is_error;
	FILE	*db_list_file = NULL;
	long	found;
	long	the_view;
	float	left_lon, right_lon, top_lat, bottom_lat;

    char            prefix[MAX_STRING_LENGTH], view_description[MAX_STRING_LENGTH],
                    user_string[MAX_STRING_LENGTH], log_string[512];
    char            view_name[MAX_STRING_LENGTH], index_view_name[MAX_STRING_LENGTH],
                    query[MAX_STRING_LENGTH], *ctime();
    char            index_file[256], *p, **pp, *hit_map;
    char           *bin_query, *sql, *select_query, *id_query;
    char            database_var[256], tempstr[256], a_char;

    long           *pi, **ppi, n_bins, *bin_list;
    long            i, j, k, id, magic_number, my_size;
    long            n_local_databases, command, n_related_views;
    long           *index_list, time(), list_length;

    double          value;
    FILE           *indfile;
    int             the_index_view, position, ref_view_index,
                    pipe_in[2], pipe_out[2];
    char           *field_pos;

    char            str[MAX_STRING_LENGTH];
    int		    lex_worked;

    n_views = 0;

    nandiv = 0.0;
    nandiv = nandiv / nandiv;
    pinfinity = 0.0;
    pinfinity = 1.0 / pinfinity;
    minfinity = (-pinfinity);


	/* preliminaries: read in known databases, skipping bad ones */

	if (argc != 7) {
		fprintf(stderr, "usage %s view query left_lon right_lon bottom_lat top_lat\n", argv[0]);
		exit(-1);
	}

	if( sscanf( argv[3], "%e", &left_lon ) != 1 ) {
		fprintf(stderr,"cant read left lon\n");
		exit(-1);
		}

	if( sscanf( argv[4], "%e", &right_lon ) != 1 ) {
		fprintf(stderr,"cant read right lon\n");
		exit(-1);
		}

	if( sscanf( argv[5], "%e", &bottom_lat ) != 1 ) {
		fprintf(stderr,"cant read bottom lat\n");
		exit(-1);
		}
	if( sscanf( argv[6], "%e", &top_lat ) != 1 ) {
		fprintf(stderr,"cant read top lat\n");
		exit(-1);
		}

	if ((db_list_file = fopen("dblist", "r")) == NULL) {
		fprintf(stderr, "cant open database list file ./dblist", argv[1]);
		exit(-1);
	}
 
	while (fscanf(db_list_file, "%s", prefix) == 1) {
		sprintf(view_description, "%s/view_description", prefix);
		if (!load_view_description(view_description, error_message)) {
			fprintf(stderr,"error loading description file\n");
			fprintf(stderr,"%s\n", error_message );
			exit(-1);
			}
	} /* end while */

	if (db_list_file != NULL)
		fclose(db_list_file);

	/* view name */
	strcpy( view_name, argv[1] );

	/* find view */
	found = FALSE;
	for (i = 0; i < n_views; i++) {
	if (!strcmp(view[i].view_name, view_name)) {
		found = TRUE;
		the_view = i;
		break;
		}
	} /* next view i */

	if (!found) {	/* non-existant view */
		sprintf(error_message, "sql: no such view <%s>", view_name);
		is_error = TRUE;
	}

	if( view[the_view].method != METHOD_BIN ) {
		fprintf(stderr,"sorry: this version of %s only works with METHOD=BIN views\n", argv[0] );
		exit(-1);
		}

	/* query */
	strcpy(query,argv[2]);

	parse_initialize("A");
	install_functions("A");
	if (!sql_parse_this_query(query, the_view, error_message)) {
		fprintf(stderr,"%s\n", error_message);
	}

	sql = (char *) malloc(256);
	sql[0] = '\0';

	/* set up 'select ... from .... where' */
	select_query = (char *) malloc(4); select_query[0]='\0';
	form_select(the_view, &select_query);
	sql = (char *) realloc(sql, strlen(sql) + strlen(select_query) + 1);
	strcat(sql, select_query);

	lex_worked=FALSE;
	lex_view=the_view;
	if( lex(query,tempstr,error_message) && strlen(tempstr)>0 ) {
	    lex_worked=TRUE;
	    sql = (char *) realloc(sql, strlen(sql) + strlen(tempstr) + 20);
            strcat(sql, LEFT_PAREN);
            strcat(sql, tempstr);
            strcat(sql, RIGHT_PAREN);
            strcat(sql, "\n");
	    }
	else if( strcmp("TRUE",query)!=0 ) {
		fprintf(stderr,"warning: couldnt convert query to sql - query reset to TRUE\n");
		}

	/* cat on the join portion of the 'where' clause */
	if (view[the_view].sql_join != NULL) {
	    if( lex_worked ) {
			strcat( sql, "and " );
			}
	    i = strlen(view[the_view].sql_join);
	    if (i > 0 && view[the_view].sql_join[i - 1] == '\n') {
		view[the_view].sql_join[i - 1] = '\0';
	    }
	    sql = (char *) realloc(sql, strlen(sql) + strlen(view[the_view].sql_join) + 20);
	    strcat(sql, LEFT_PAREN);
	    strcat(sql, view[the_view].sql_join);
	    strcat(sql, RIGHT_PAREN);
	    strcat(sql, "\n");
	}

	form_bin_query_lat(left_lon, right_lon, bottom_lat, top_lat, &bin_query, the_view);
	sql = (char *) realloc(sql, strlen(sql) + strlen(bin_query) + 1);
	strcat(sql, bin_query);
	sql = (char *) realloc(sql, strlen(sql) + 10);

	doitall( &view[the_view], sql );
	exit(0);

}

doitall( v, sql )
struct view_rec *v;
char *sql;
{
	printf("VIEW NAME: <%s>\n", v->view_name );
	printf("%s\n", sql );
	}


form_select(the_view, select_query)
    int             the_view;
    char          **select_query;
{
    int             i, j, found, first;
    char            phrase[12 * SHORT_STRING_LENGTH + 2];

    strcpy(phrase, "SET SQLN OFF;\n");
    strcat(phrase, "SET RECSEP OFF;\n");
    strcat(phrase, "SET EMBEDDED OFF;\n");
    strcat(phrase, "SET FEEDBACK OFF;\n");
    strcat(phrase, "SET FLUSH OFF;\n");
    strcat(phrase, "SET NEWPAGE 1;\n");
    strcat(phrase, "SET LINESIZE 500;\n");
    strcat(phrase, "SET PAGESIZE 100;\n");
    strcat(phrase, "SET SPACE 1;\n");
    strcat(phrase, "SET WRAP ON;\n");
    strcat(phrase, "SET TAB OFF;\n");
    strcat(phrase, "SET UNDERLINE \\;\n");


    strcat(phrase, "select\n");
    *select_query = (char *) realloc(*select_query, strlen(*select_query) + strlen(phrase) + 1);
    strcpy(*select_query, phrase);

    sprintf(phrase, "    %s,\n",  view[the_view].file_field );
    *select_query = (char *) realloc(*select_query, strlen(*select_query) + strlen(phrase) + 1);
    strcat(*select_query, phrase);

    strcpy(phrase, "from\n");
    *select_query = (char *) realloc(*select_query, strlen(*select_query) + strlen(phrase) + 1);
    strcat(*select_query, phrase);

    first = TRUE;
    for (i = 0; i < view[the_view].n_fields; i++) {
	found = FALSE;
	for (j = 0; j < i; j++) {
	    if (!strcmp(view[the_view].f[i].database_table_name, view[the_view].f[j].database_table_name)) {
		found = TRUE;
		break;
	    }
	}
	if (!found) {
	    if (first) {
		first = FALSE;
		sprintf(phrase, "%s.%s", view[the_view].f[i].database_name, view[the_view].f[i].database_table_name);
	    }
	    else {
		sprintf(phrase, ", %s.%s", view[the_view].f[i].database_name, view[the_view].f[i].database_table_name);
	    }
	    *select_query = (char *) realloc(*select_query, strlen(*select_query) + strlen(phrase) + 1);
	    strcat(*select_query, phrase);
	}
    }

    if (strlen(view[the_view].extra_tables) != 0) {
	sprintf(phrase, ", %s", view[the_view].extra_tables);
	*select_query = (char *) realloc(*select_query, strlen(*select_query) + strlen(phrase) + 1);
	strcat(*select_query, phrase);
    }
    strcpy(phrase, " where\n");
    *select_query = (char *) realloc(*select_query, strlen(*select_query) + strlen(phrase) + 1);
    strcat(*select_query, phrase);

}

form_bin_query_lat(left_lon, right_lon, bottom_lat, top_lat, bin_query, the_view)
    double left_lon, right_lon, bottom_lat, top_lat;
    char          **bin_query;
    int             the_view;
{
    int		    lat_field, lon_field;
    char            phrase[256];

    lat_field = view[the_view].lat_field;
    lon_field = view[the_view].lon_field;

    *bin_query = (char *) malloc( 10 );

    strcpy( (*bin_query), LEFT_PAREN );

	if( !view[the_view].longitude_is_neg || right_lon<180.0 ) { /*dont have to worry about negative lons */
    		sprintf(phrase, "%s %s.%s.%s between %f and %f and %s.%s.%s between %f and %f %s",
		LEFT_PAREN,
		view[the_view].f[lon_field].database_name,
       		view[the_view].f[lon_field].database_table_name,
        	view[the_view].f[lon_field].database_field_name,
		left_lon,
		right_lon,
        	view[the_view].f[lat_field].database_name,
       	 	view[the_view].f[lat_field].database_table_name,
        	view[the_view].f[lat_field].database_field_name,
		bottom_lat,
		top_lat,
		RIGHT_PAREN );
		}
	else if( view[the_view].longitude_is_neg && left_lon>=180.0 ) { /* only one box, which is negative */
		sprintf(phrase, "%s %s.%s.%s between %f and %f and %s.%s.%s between %f and %f %s",
		LEFT_PAREN,
		view[the_view].f[lon_field].database_name,
                view[the_view].f[lon_field].database_table_name,
                view[the_view].f[lon_field].database_field_name,
		left_lon-360.0,
		right_lon-360.0,
                view[the_view].f[lat_field].database_name,
                view[the_view].f[lat_field].database_table_name,
                view[the_view].f[lat_field].database_field_name,
		bottom_lat,
		top_lat,
		RIGHT_PAREN );
                }
	else if( view[the_view].longitude_is_neg && left_lon<180.0 ) { /* two boxes */
		/* first box on positive side, has right at 180 */
		sprintf(phrase, "%s %s.%s.%s between %f and %f and %s.%s.%s between %f and %f %s",
		LEFT_PAREN,
		view[the_view].f[lon_field].database_name,
                view[the_view].f[lon_field].database_table_name,
                view[the_view].f[lon_field].database_field_name,
		left_lon,
		180.0,
                view[the_view].f[lat_field].database_name,
                view[the_view].f[lat_field].database_table_name,
                view[the_view].f[lat_field].database_field_name,
		bottom_lat,
		top_lat,
		RIGHT_PAREN );

		*bin_query = (char *) realloc(*bin_query, strlen(*bin_query) + strlen(phrase) + 4);
		strcat(*bin_query, phrase);
	   	strcat(*bin_query, " or\n");

		/* second box on negative side, has left at -180 */
		sprintf(phrase, "%s %s.%s.%s between %f and %f and %s.%s.%s between %f and %f %s",
		LEFT_PAREN,
		view[the_view].f[lon_field].database_name,
                view[the_view].f[lon_field].database_table_name,
                view[the_view].f[lon_field].database_field_name,
		-180.0,
		right_lon-360.0,
                view[the_view].f[lat_field].database_name,
                view[the_view].f[lat_field].database_table_name,
                view[the_view].f[lat_field].database_field_name,
		bottom_lat,
		top_lat,
		RIGHT_PAREN );
	}

	*bin_query = (char *) realloc(*bin_query, strlen(*bin_query) + strlen(phrase) + 4);
	strcat(*bin_query, phrase);
	strcat(*bin_query, RIGHT_PAREN);
	strcat(*bin_query, ";\n");
}


int 
load_view_description(file, error_message)
    char           *file;
    char           *error_message;
{
    int             i, j, k, l, m, id, found, nwords, view_open;
    char            line[256], copy_of_line[256], *word[128], str[256];
    FILE           *f;

    if ((f = fopen(file, "r")) == NULL) {
	sprintf(error_message, "cant open <%s> to read description\n", file);
	return (FALSE);
    }

    view_open = FALSE;
    /* initialize */
    for (i = n_views; i < MAX_VIEWS; i++) {
	sprintf(str, "view_%d");
	strcpy(view[i].view_name, str);
	view[i].method = METHOD_BIN;
	strcpy(view[i].description, "view description not available");
	view[i].sql_server[0] = '\0';
	view[i].shell[0] = '\0';
	view[i].reformatter[0] = '\0';
	view[i].file_field[0] = '\0';
	view[i].longitude_is_neg = FALSE;
	view[i].hitmap_path[0] = '\0';
	view[i].bin_field[0] = '\0';
	view[i].extra_tables[0] = '\0';
	view[i].n_fields = 0;
	view[i].sql_join = NULL;
	view[i].n_ids = 0;
	for (j = 0; j < MAX_IDS; j++)
	    view[i].sql_id[j] = NULL;
	view[i].n_related = 0;
    }


    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */

	if (!strcmp(word[0], "VIEW_DEFINITION")) {
	    view_open = TRUE;
	    if ((++n_views) > MAX_VIEWS) {
		strcpy(error_message, "too many views");
		fclose(f);
		return (FALSE);
	    }
	    if (nwords != 4) {
		strcpy(error_message, "wrong number of words for VIEW_DEFINITION <%s>", word[1]);
		fclose(f);
		return (FALSE);
	    }
	    strcpy(view[n_views - 1].view_name, word[1]);
	    /* check if view name not already defined */
	    for (i = 0; i < (n_views - 2); i++) {
		if (!strcmp(view[i].view_name, view[n_views - 1].view_name)) {
		    sprintf(error_message, "view <%s> multiply defined", view[n_views - 1].view_name);
		    fclose(f);
		    return (FALSE);
		}
	    }

	    if (!strcmp(word[3], "ID")) {
		view[n_views - 1].method = METHOD_RECORD_ID;
	    }
	    else if (!strcmp(word[3], "BIN")) {
		view[n_views - 1].method = METHOD_BIN;
	    }
	    else {
		strcpy(error_message, "VIEW_DEFINITION <%s> has bad method <%s>", word[1], word[3]);
		fclose(f);
		return (FALSE);
	    }

	    while (fgets(line, 256, f) != NULL) {

		if (line[0] == '#')
		    continue;	/* comment lines allowed */
		strcpy(copy_of_line, line);
		sbreak(copy_of_line, word, &nwords);
		if (nwords < 1)
		    continue;	/* blank lines allowed */

		if (!strcmp(word[0], "DESCRIPTION")) {
		    if (!read_description(&view[n_views - 1], f, error_message))
			return (FALSE);
		}
		else if (!strcmp(word[0], "SQL_SERVER")) {
		    if (!read_sql_server(&view[n_views - 1], f, error_message))
			return (FALSE);
		}
		else if (!strcmp(word[0], "SHELL")) {
		    if (!read_shell(&view[n_views - 1], f, error_message))
			return (FALSE);
		}
		else if (!strcmp(word[0], "REFORMATTER")) {
		    if (!read_reformatter(&view[n_views - 1], f, error_message))
			return (FALSE);
		}
		else if (!strcmp(word[0], "FILENAME_FIELD")) {
		    if (!read_file_name(&view[n_views - 1], f, error_message))
			return (FALSE);
		}
                else if (!strcmp(word[0], "LONGITUDE_RANGE")) {
                    if (!read_longitude_range(&view[n_views - 1], f, error_message))
                        return (FALSE);
                }
		else if (!strcmp(word[0], "HITMAP_PATH")) {
		    if (!read_hitmap_path(&view[n_views - 1], f, error_message))
			return (FALSE);
		}
		else if (!strcmp(word[0], "BIN_FIELD")) {
		    if (!read_bin_field(&view[n_views - 1], f, error_message))
			return (FALSE);
		}
		else if (!strcmp(word[0], "TABLES_NOT_IN_SCHEMA")) {
		    if (!read_extra_tables(&view[n_views - 1], f, error_message))
			return (FALSE);
		}
		else if (!strcmp(word[0], "SCHEMA")) {
		    if (!read_schema(&view[n_views - 1], f, error_message))
			return (FALSE);
		}
		else if (!strcmp(word[0], "SQL_JOIN")) {
		    if (!read_sql_join(&view[n_views - 1], f, error_message))
			return (FALSE);
		}
		else if (!strcmp(word[0], "SQL_ID")) {
		    if (!read_sql_id(&view[n_views - 1], f, word[1], nwords, error_message))
			return (FALSE);
		}
		else if (!strcmp(word[0], "VIEWS_TO_WHICH_THIS_VIEW_IS_RELATED")) {
		    if (!read_views_related(&view[n_views - 1], f, error_message))
			return (FALSE);
		}
		else if (!strcmp(word[0], "END")) {
		    break;
		}
		else {
		    sprintf(error_message, "VIEW <%s> has bad keyword <%s>", view[n_views - 1].view_name,
			    word[0]);
		    fclose(f);
		    return (FALSE);
		}
	    }			/* end while */
	}			/* end if */
    }				/* end while */

    fclose(f);

    /* now check for consistancy */
    for (i = 0; i < n_views; i++) {

	/* fielname field must be present and non-blank */
	if( strlen( view[i].file_field ) == 0 ) {
		sprintf(error_message, "view %d must have FILENAME_FIELD on non-zero length\n", i );
		return(FALSE);
		}

	/* id views must have atleast one non-null sql_id field */
	found = FALSE;
	for (j = 0; j < MAX_IDS; j++) {
	    if (view[i].sql_id[j] != NULL) {
		found = TRUE;
		break;
	    }
	}
	if ((view[i].method == METHOD_RECORD_ID) && (!found)) {
	    sprintf(error_message, "view <%s> by ID must have SQL_ID", view[i].view_name);
	    return (FALSE);
	}

	/* views referenced must exist */
	for (j = 0; j < view[i].n_related; j++) {
	    found = FALSE;
	    for (k = 0; k < n_views; k++) {
		if (!strcmp(view[i].r[j].view_name, view[k].view_name)) {
		    view[i].r[j].view_index = k;
		    found = TRUE;
		    break;
		}
	    }
	    if (!found) {
		sprintf(error_message, "view <%s> reference nonexistant view <%s>", view[i].view_name,
			view[i].r[j].view_name);
		return (FALSE);
	    }
	}

	/* ids referenced must exist */
	for (j = 0; j < view[i].n_related; j++) {
	    if (view[i].sql_id[view[i].r[j].id] == NULL) {
		sprintf(error_message, "view <%s> reference nonexistant sql_id %d",
			view[i].view_name,
			view[i].r[j].id);
		return (FALSE);
	    }
	}

	/* view_fields in referenced views must exist */
	for (j = 0; j < view[i].n_related; j++) {
	    k = view[i].r[j].view_index;
	    for (l = 0; l < view[i].r[j].n_variables; l++) {
		found = FALSE;
		for (m = 0; m < view[k].n_fields; m++) {
		    if (!strcmp(view[i].r[j].v[l].view_field_name, view[k].f[m].name)) {
			view[i].r[j].v[l].field_index = m;
			found = TRUE;
			break;
		    }
		}
		if (!found) {
		    sprintf(error_message,
			    "view <%s> references nonexistant field <%s> in view <%s>",
			    view[i].view_name, view[i].r[j].v[l].view_field_name, view[k].view_name);
		    return (FALSE);
		}
	    }
	}

	/* variables in sql_id must be defined for all referenced views */
	if (view[i].method == METHOD_RECORD_ID) {
	    for (j = 0; j < MAX_IDS; j++)
		find_variables(&n_vars[j], view[i].sql_id[j], j);
	    for (j = 0; j < view[i].n_related; j++) {
		id = view[i].r[j].id;
		for (k = 0; k < n_vars[id]; k++) {
		    found = FALSE;
		    for (l = 0; l < view[i].r[j].n_variables; l++) {
			if (!strcmp(vlist[id][k], view[i].r[j].v[l].variable_name)) {
			    found = TRUE;
			    break;
			}
		    }
		    if (!found) {
			sprintf(error_message, "view <%s> doesnt define variable <%s> for reference view <%s>",
				view[i].view_name, vlist[id][k], view[i].r[j].view_name);
			return (FALSE);
		    }
		}
	    }
	}

	/* compute field offsets */
	if (view[i].n_fields > 0) {
	    view[i].f[0].offset = 0;
	    for (j = 1; j < view[i].n_fields; j++) {
		view[i].f[j].offset = view[i].f[j - 1].offset + view[i].f[j - 1].length;
	    }
	    view[i].buffer_length = view[i].f[view[i].n_fields - 1].offset + view[i].f[view[i].n_fields - 1].length;
	}

	/* find lat field */
	if (view[i].method == METHOD_BIN) {
	    found = FALSE;
	    for (j = 0; j < view[i].n_fields; j++) {
		if (!strcmp(view[i].f[j].name, "lat")) {
		    found = TRUE;
		    view[i].lat_field = j;
		    break;
		}
	    }
	    if (!found) {
		sprintf(error_message, "view <%s> needs a lat field", view[i].view_name);
		return (FALSE);
	    }
	}
	/* find lon fields */
	if (view[i].method == METHOD_BIN) {
	    found = FALSE;
	    for (j = 0; j < view[i].n_fields; j++) {
		if (!strcmp(view[i].f[j].name, "lon")) {
		    found = TRUE;
		    view[i].lon_field = j;
		    break;
		}
	    }
	    if (!found) {
		sprintf(error_message, "view <%s> needs a lon field", view[i].view_name);
		return (FALSE);
	    }
	}

    }
    return (TRUE);
}				/* end subroutine */

find_variables(n, str, id)
    int            *n;
    char           *str;
    int             id;
{
    char           *c, s[2];
    int             vopen;

    if (str == NULL) {
	(*n) = 0;
	return;
    }

    vopen = FALSE;
    (*n) = 0;
    for (c = str; (*c) != '\0'; c++) {
	if (((*c) == ' ') || ((*c) == '\t') || ((*c) == '\n') || ((*c) == '\r')) {
	    vopen = FALSE;
	    continue;
	}
	else if ((*c) == '$') {
	    vopen = TRUE;
	    (*n)++;
	    s[0] = (*c);
	    s[1] = '\0';
	    vlist[id][(*n) - 1][0] = '\0';
	    strcat(vlist[id][(*n) - 1], s);
	}
	else if (vopen) {
	    s[0] = (*c);
	    s[1] = '\0';
	    strcat(vlist[id][(*n) - 1], s);
	}
    }
}


read_description(v, f, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *error_message;
{
    int             nwords;
    char            line[256], copy_of_line[256], *word[128], str[256];
    v->description[0] = '\0';
    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */

	if (nwords != 1) {
	    sprintf(error_message, "bad description in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	if (!strcmp(word[0], "END"))
	    break;

	strcat(v->description, word[0]);
	strcat(v->description, "\n");
    }
    return (TRUE);

}

read_sql_server(v, f, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *error_message;
{
    char            line[256], copy_of_line[256], *word[128], str[256];
    int             count;
    int             nwords;


    count = 0;
    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */

	if (nwords != 1) {
	    sprintf(error_message, "bad sql_server in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	if (!strcmp(word[0], "END"))
	    break;

	strcpy(v->sql_server, word[0]);
	count++;
    }

    if (count != 1) {
	sprintf(error_message, "wrong number of sql_servers in view <%s>", v->view_name);
	fclose(f);
	return (FALSE);
    }
    return (TRUE);

}

read_shell(v, f, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *error_message;
{
    char            line[256], copy_of_line[256], *word[128], str[256];
    int             count;
    int             nwords;


    count = 0;
    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */

	if (nwords != 1) {
	    sprintf(error_message, "bad shell in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	if (!strcmp(word[0], "END"))
	    break;

	strcpy(v->shell, word[0]);
	count++;
    }

    if (count != 1) {
	sprintf(error_message, "wrong number of shells in view <%s>", v->view_name);
	fclose(f);
	return (FALSE);
    }
    return (TRUE);

}

read_reformatter(v, f, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *error_message;
{
    char            line[256], copy_of_line[256], *word[128], str[256];
    int             count;
    int             nwords;


    count = 0;
    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */

	if (nwords != 1) {
	    sprintf(error_message, "bad reformatter in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	if (!strcmp(word[0], "END"))
	    break;

	strcpy(v->reformatter, word[0]);
	count++;
    }

    if (count != 1) {
	sprintf(error_message, "wrong number of reformatterss in view <%s>", v->view_name);
	fclose(f);
	return (FALSE);
    }
    return (TRUE);

}

read_file_name(v, f, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *error_message;
{
    char            line[256], copy_of_line[256], *word[128], str[256];
    int             count;
    int             nwords;


    count = 0;
    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */

	if (nwords != 1) {
	    sprintf(error_message, "bad filename_field in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	if (!strcmp(word[0], "END"))
	    break;

	strcpy(v->file_field, word[0]);
	count++;
    }

    if (count != 1) {
	sprintf(error_message, "wrong number of filename_fields in view <%s>", v->view_name);
	fclose(f);
	return (FALSE);
    }
    return (TRUE);

}
read_longitude_range(v, f, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *error_message;
{
    char            line[256], copy_of_line[256], *word[128], str[256];
    int             count;
    int             nwords;


    count = 0;
    while (fgets(line, 256, f) != NULL) {

        if (line[0] == '#')
            continue;           /* comment lines allowed */
        strcpy(copy_of_line, line);
        sbreak(copy_of_line, word, &nwords);
        if (nwords < 1)
            continue;           /* blank lines allowed */

        if (nwords != 1) {
            sprintf(error_message, "bad longitude range keyword in view <%s>", v->view_name);
            fclose(f);
            return (FALSE);
        }
        if (!strcmp(word[0], "END"))
            break;

        if( !strcmp(word[0],"IS_NEGATIVE") ) {
		v->longitude_is_neg=TRUE;
		}
	else if( !strcmp(word[0],"IS_POSITIVE") ) {
		v->longitude_is_neg=FALSE;
		}
	else {
		sprintf(error_message, "bad longitude range keyword <%s>\n", word[0] );
		fclose(f);
		return(FALSE);
		}
        count++;
    }

    if (count != 1) {
        sprintf(error_message, "wrong number of longitude range keywords in view <%s>", v->view_name);
        fclose(f);
        return (FALSE);
    }
    return (TRUE);
 
}



read_hitmap_path(v, f, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *error_message;
{
    char            line[256], copy_of_line[256], *word[128], str[256];
    int             count;
    int             nwords;


    count = 0;
    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */

	if (nwords != 1) {
	    sprintf(error_message, "bad hitmap path in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	if (!strcmp(word[0], "END"))
	    break;

	strcpy(v->hitmap_path, word[0]);
	count++;
    }

    if (count != 1) {
	sprintf(error_message, "wrong number of hitmap paths in view <%s>", v->view_name);
	fclose(f);
	return (FALSE);
    }
    return (TRUE);
}

read_bin_field(v, f, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *error_message;
{
    char            line[256], copy_of_line[256], *word[128], str[256];
    int             count;
    int             nwords;


    count = 0;
    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */

	if (nwords != 1) {
	    sprintf(error_message, "bad bin field in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	if (!strcmp(word[0], "END"))
	    break;

	strcpy(v->bin_field, word[0]);
	count++;
    }

    if (count != 1) {
	sprintf(error_message, "wrong number of bin fields in view <%s>", v->view_name);
	fclose(f);
	return (FALSE);
    }
    return (TRUE);
}

read_extra_tables(v, f, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *error_message;
{
    char            line[256], copy_of_line[256], *word[128], str[256];
    int             count;
    int             nwords;


    count = 0;
    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */

	if (nwords != 1) {
	    sprintf(error_message, "bad TABLES_NOT_IN_SCHEMA in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	if (!strcmp(word[0], "END"))
	    break;

	strcpy(v->extra_tables, word[0]);
	count++;
    }

    if (count != 1) {
	sprintf(error_message, "bad TABLES_NOT_IN_SCHEMA in view <%s>", v->view_name);
	fclose(f);
	return (FALSE);
    }
    return (TRUE);
}

read_schema(v, f, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *error_message;
{
    char            line[256], copy_of_line[256], *word[128], str[256];
    int             i, is_string;
    int             nwords;

    v->n_fields = 0;
    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */
	if (!strcmp(word[0], "END"))
	    break;

	v->n_fields++;

	if ((v->n_fields + 2) > MAX_FIELDS) {
	    sprintf(error_message, "too many fields in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	if (nwords < 6) {
	    sprintf(error_message, "bad SCHEMA in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	if (!strcmp(word[0], "END"))
	    break;
	is_string = 0;

	strcat(v->f[v->n_fields - 1].name, word[0]);
	if (!strcmp(word[1], "SHORT") || !strcmp(word[1], "short")) {
	    v->f[v->n_fields - 1].type = SHORT_TYPE;
	    v->f[v->n_fields - 1].length = sizeof(short);
	}
	if (!strcmp(word[1], "LONG") || !strcmp(word[1], "long") ||
	    !strcmp(word[1], "INT") || !strcmp(word[1], "int")) {
	    v->f[v->n_fields - 1].type = LONG_TYPE;
	    v->f[v->n_fields - 1].length = sizeof(long);
	}
	if (!strcmp(word[1], "FLOAT") || !strcmp(word[1], "float")) {
	    v->f[v->n_fields - 1].type = FLOAT_TYPE;
	    v->f[v->n_fields - 1].length = sizeof(float);
	}
	if (!strcmp(word[1], "DOUBLE") || !strcmp(word[1], "double")) {
	    v->f[v->n_fields - 1].type = DOUBLE_TYPE;
	    v->f[v->n_fields - 1].length = sizeof(double);
	}
	if (!strcmp(word[1], "STR") || !strcmp(word[1], "str") ||
	    !strcmp(word[1], "STRING") || !strcmp(word[1], "string")) {
	    is_string = 1;
	    v->f[v->n_fields - 1].type = STRING_TYPE;
	    if (sscanf(word[2], "%d", &i) != 1 || i < 2 || i > MAX_STRING_LENGTH) {
		sprintf(error_message, "bad string length in view <%s>", v->view_name);
		fclose(f);
		return (FALSE);
	    }
	    v->f[v->n_fields - 1].length = i;
	}
	if (nwords != (6 + is_string)) {
	    sprintf(error_message, "bad SCHEMA in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	strcat(v->f[v->n_fields - 1].database_name, word[2 + is_string]);
	strcat(v->f[v->n_fields - 1].database_table_name, word[3 + is_string]);
	strcat(v->f[v->n_fields - 1].database_field_name, word[4 + is_string]);
	strcat(v->f[v->n_fields - 1].comment, word[5 + is_string]);
    }
    return (TRUE);
}

read_sql_join(v, f, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *error_message;
{
    int             nwords;

    char            line[256], copy_of_line[256], *word[128], str[256];
    int             size;

    size = 2;
    v->sql_join = malloc(size);
    v->sql_join[0] = '\0';

    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */

	if (nwords != 1) {
	    sprintf(error_message, "bad sql join in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	if (!strcmp(word[0], "END"))
	    break;

	size += strlen(word[0]) + 2;
	v->sql_join = realloc(v->sql_join, size);
	strcat(v->sql_join, word[0]);
	strcat(v->sql_join, "\n");
    }
    return (TRUE);
}

read_sql_id(v, f, tag, nw, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *tag;
    int             nw;
    char           *error_message;
{
    char            line[256], copy_of_line[256], *word[128], str[256];
    int             size;
    int             nwords, id;

    if (nw == 1) {
	id = 0;
    }
    else {
	if (sscanf(tag, "%d", &id) != 1) {
	    sprintf(error_message, "bad sql_id number in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
    }

    if (id < 0 || id > MAX_IDS - 1) {
	sprintf(error_message, "bad sql_id number in view <%s>", v->view_name);
	fclose(f);
	return (FALSE);
    }

    size = 2;
    v->sql_id[id] = malloc(size);
    v->sql_id[id][0] = '\0';

    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */

	if (nwords != 1) {
	    sprintf(error_message, "bad sql_id in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}
	if (!strcmp(word[0], "END"))
	    break;

	size += strlen(word[0]) + 2;
	v->sql_id[id] = realloc(v->sql_id[id], size);
	strcat(v->sql_id[id], word[0]);
	strcat(v->sql_id[id], "\n");
    }
    return (TRUE);
}

read_views_related(v, f, error_message)
    struct view_rec *v;
    FILE           *f;
    char           *error_message;
{
    char            line[256], copy_of_line[256], *word[128], str[256],
                    s2[256];
    int             nwords;

    v->n_related = 0;

    while (fgets(line, 256, f) != NULL) {

	if (line[0] == '#')
	    continue;		/* comment lines allowed */
	strcpy(copy_of_line, line);
	sbreak(copy_of_line, word, &nwords);
	if (nwords < 1)
	    continue;		/* blank lines allowed */

	if ((nwords == 1) && !strcmp(word[0], "END"))
	    break;
	if (nwords != 2 && nwords != 5) {
	    sprintf(error_message, "bad related views in view <%s>", v->view_name);
	    fclose(f);
	    return (FALSE);
	}

	v->n_related++;
	strcpy(v->r[v->n_related - 1].view_name, word[1]);
	v->r[v->n_related - 1].n_variables = 0;

	if (nwords == 2) {
	    v->r[v->n_related - 1].id = 0;
	}
	else if (nwords == 5) {
	    if (sscanf(word[4], "%d", &(v->r[v->n_related - 1].id)) != 1
		|| v->r[v->n_related - 1].id < 0
		|| v->r[v->n_related - 1].id > MAX_IDS - 1) {
		sprintf(error_message, "bad related views in view <%s>", v->view_name);
		fclose(f);
		return (FALSE);
	    }
	}

	while (fgets(line, 256, f) != NULL) {

	    if (line[0] == '#')
		continue;	/* comment lines allowed */
	    strcpy(copy_of_line, line);
	    sbreak(copy_of_line, word, &nwords);
	    if ((nwords == 1) && !strcmp(word[0], "END"))
		break;

	    if (nwords == 2) {
		s2[0] = '\0';
	    }
	    else if (nwords == 3) {
		strcpy(s2, word[2]);
	    }
	    else {
		sprintf(error_message, "bad related view variables in view <%s>", v->view_name);
		fclose(f);
		return (FALSE);
	    }

	    v->r[v->n_related - 1].n_variables++;
	    strcpy(v->r[v->n_related - 1].v[v->r[v->n_related - 1].n_variables - 1].variable_name, word[0]);
	    strcpy(v->r[v->n_related - 1].v[v->r[v->n_related - 1].n_variables - 1].view_field_name, word[1]);
	    strcpy(v->r[v->n_related - 1].v[v->r[v->n_related - 1].n_variables - 1].format, s2);
	}

    }
    return (TRUE);
}

long 
sbreak(in, out, n)
    char           *in, *out[];
    long           *n;
{
    char            c;
    long            i, l, quote, word, first;

    *n = 0;
    l = strlen(in);
    quote = FALSE;
    word = FALSE;
    for (i = 0; i < l; i++) {
	c = in[i];
	if (c == '"') {
	    if (!quote) {
		first = TRUE;
		word = TRUE;
		quote = TRUE;
	    }
	    else if (quote) {
		quote = FALSE;
		word = FALSE;
		in[i] = '\0';
	    }
	}
	else if ((!quote) && (c == ' ' || c == '\t' || c == '\n')) {
	    in[i] = '\0';
	    word = FALSE;
	}
	else {
	    if (!word) {
		word = TRUE;
		first = TRUE;
	    }
	    if (first) {
		out[(*n)++] = &in[i];
		first = FALSE;
	    }
	}
    }				/* end for */
}

double 
par_time(x)
    double         *x;
{
    double          y;
    long            time;
    struct tm      *gmtime(), *tm;

    time = (long) (*x);
    tm = gmtime(&time);

    y = tm->tm_sec + 100 * tm->tm_min + 10000 * tm->tm_hour;

    return (y);
}

double 
par_date(x)
    double         *x;
{
    double          y;
    long            time;
    struct tm      *gmtime(), *tm;

    time = (long) (*x);
    tm = gmtime(&time);

    y = tm->tm_mday + 100 * (tm->tm_mon + 1) + 10000 * tm->tm_year;

    errno = 0;
    return (y);
}

install_functions(parser)
    char           *parser;
{
    parse_installfunc(parser, "epochtodate", par_date, PARSE_NUM, 1, PARSE_NUM);
    parse_installfunc(parser, "epochtotime", par_time, PARSE_NUM, 1, PARSE_NUM);
}

long 
sql_parse_this_query(str, the_view, error_message)
    char           *str;
    int             the_view;
    char           *error_message;
{
    long            value, error, i, j, found, symbol;
    char           *d, *data_list, *n, name_buf[100];

    if (!strcmp(str, "TRUE")) {
	query_is_true = TRUE;
	return (TRUE);
    }
    else
	query_is_true = FALSE;

    error = PARSE_NOERR;
    if ((error = parse("A", str, &data_list)) != PARSE_NOERR) {
	sprintf(error_message, "sql: cant parse query <%s>", str);
	return (FALSE);
    }

    /* count up the number of variables */
    sql_par_n_var = 0;
    for (d = data_list; *d != '\0'; d++) {
	for (n = name_buf;; n++, d++)
	    if ((*n = *d) == '\0')
		break;
	sql_par_n_var++;
    }

    if (sql_par_n_var == 0) {
	sql_par_table = NULL;
    }
    else {
	sql_par_table = (struct sql_ascii_rec *) malloc(sql_par_n_var * sizeof(struct sql_ascii_rec));
    }

    /* match the variables with the fields of the record */
    j = -1;
    for (d = data_list; *d != '\0'; d++) {
	j++;

	for (n = name_buf;; n++, d++)
	    if ((*n = *d) == '\0')
		break;

	found = FALSE;
	for (i = 0; i < view[the_view].n_fields; i++) {
	    if (strcmp(name_buf, view[the_view].f[i].name) == 0) {
		sql_par_table[j].offset = view[the_view].f[i].offset;
		sql_par_table[j].type = view[the_view].f[i].type;
		strcpy(sql_par_table[j].name, view[the_view].f[i].name);
		found = TRUE;
		break;
	    }
	}

	if (!found) {
	    sprintf(error_message, "smd: symbol <%s> in parse expression not found among fields of view", name_buf);
	    return (FALSE);
	}
    }

    return (TRUE);
}

long 
evaluate_query(buffer, the_view, the_value, error_message)
    char           *buffer;
    int             the_view;
    double         *the_value;
    char           *error_message;
{
    long            i, error1, error2, type;
    char           *p, *d, *name, *result;
    double          value;
    char           *data_list;

    if (query_is_true) {
	*the_value = (double) 1.0;
	return (TRUE);
    }

    for (i = 0; i < sql_par_n_var; i++) {
	d = (char *) buffer;
	switch (sql_par_table[i].type) {

	  case SHORT_TYPE:
	    name = sql_par_table[i].name;
	    p = (char *) (((char *) d) + sql_par_table[i].offset);
	    error1 = parse_vartype("A", name, PARSE_NUM);
	    bcopy((short *) p, &my_short, sizeof(short));
	    value = (double) my_short;
	    error2 = parse_varset("A", name, &value);
	    break;

	  case LONG_TYPE:
	    name = sql_par_table[i].name;
	    p = (char *) (((char *) d) + sql_par_table[i].offset);
	    error1 = parse_vartype("A", name, PARSE_NUM);
	    bcopy((int *) p, &my_long, sizeof(long));
	    value = (double) my_long;
	    error2 = parse_varset("A", name, &value);
	    break;

	  case FLOAT_TYPE:
	    name = sql_par_table[i].name;
	    p = (char *) (((char *) d) + sql_par_table[i].offset);
	    error1 = parse_vartype("A", name, PARSE_NUM);
	    bcopy((float *) p, &my_float, sizeof(float));
	    value = (double) my_float;
	    error2 = parse_varset("A", name, &value);
	    break;

	  case DOUBLE_TYPE:
	    name = sql_par_table[i].name;
	    p = (char *) (((char *) d) + sql_par_table[i].offset);
	    error1 = parse_vartype("A", name, PARSE_NUM);
	    bcopy((double *) p, &my_double, sizeof(double));
	    value = (double) my_double;
	    error2 = parse_varset("A", name, &value);
	    break;

	  case STRING_TYPE:
	    name = sql_par_table[i].name;
	    p = (char *) (((char *) d) + sql_par_table[i].offset);
	    error1 = parse_vartype("A", name, PARSE_STR);
	    error2 = parse_varset("A", name, (char *) p);
	    break;
	}			/* end switch */

	if (error1 != PARSE_NOERR || error2 != PARSE_NOERR) {
	    sprintf(error_message, "sql: parse error (cant set variable)");
	    return (FALSE);
	}
    }				/* end for */

    error1 = parse_execute("A", &result, &type);
    if (error1 != PARSE_NOERR) {
	sprintf(error_message, "sql: parser execution error");
	return (FALSE);
    }

    if (type == PARSE_NUM)
	value = *((double *) result);
    else {
	sprintf(error_message, "sql: parse expression not a DOUBLE");
	return (FALSE);
    }

    *the_value = value;
    return (TRUE);
}


char *field_lookup( p )
char *p;
{
	int i, found;

	found=FALSE;
	for(i=0; i< view[lex_view].n_fields; i++ ) {
		if( !strcmp(view[lex_view].f[i].name,p) ) {
			found=TRUE;
			break;
			}
		}
	if( found ) {
		sprintf( lex_field_lookup_buf, "%s.%s.%s", view[lex_view].f[i].database_name,
		view[lex_view].f[i].database_table_name, view[lex_view].f[i].database_field_name );
		return(lex_field_lookup_buf);
		}
	else {
		return(NULL);
		}
	}

