#include <stdio.h>
#include <math.h>
#include <setjmp.h>
#ifdef DBVISTA
#include "dbvista_sql.h"
#endif

# define FALSE 0
# define TRUE 1

# define NUMBER 257
# define STRING 258
# define VAR_UNDEF 260
# define BLTIN_0 264
# define BLTIN_1 265
# define BLTIN_2 266
# define BLTIN_3 267
# define BLTIN_4 268
# define BLTIN_5 269
# define BLTIN_6 270
# define BLTIN_7 271
# define BLTIN_8 272
# define BLTIN_9 273
# define BLTIN_10 274
# define EOS 275
# define OR 276
# define AND 277
# define EQ 278
# define NE 279
# define GT 280
# define GE 281
# define LT 282
# define LE 283
# define NOT 285
# define SQL_ERROR 300

#define RIGHT_PAREN     ")"
#define LEFT_PAREN      "("

char sql_str[256];
char *sql_str_ptr;
double sql_number;
char sql_buf[256];

jmp_buf sql_env;
char sql_error_message[256];

int n_constants=4;

char *constant_symbol[] = {
	"PI",
	"E",
	"RTOD",
	"DTOR",
	};

char *constant_value[] = {
	"3.141592654",
	"2.718281828",
	"57.29577951",
	"0.017453292",
	};


/*
main(argc,argv)
int argc;
char *argv[];
{
	char out[1024], error_message[256];

	if( lex(argv[1], out, error_message) ) {
		printf("%s\n", out);
		exit(0);
		}
	else {
		printf("error: %s\n", error_message);
		exit(-1);
		}

	}
*/

int lex( in, out, error_message )
char *in, *out, *error_message;
{
	int op, len, comparison_found;
        char buf[256], *c, *lookup();

	out[0]='\0';
	sql_str_ptr = sql_str;
	strcpy(sql_str, in );
	len=0;
	comparison_found=FALSE;
	while( (op=sql_lex()) != EOS ) {
		switch( op ) {
			case NUMBER:
				if( sql_number < 1.e-3  && sql_number != 0.0) {
					sprintf(buf,"%.20f",sql_number);
					}
				else {
					sprintf(buf,"%.6f",sql_number);
					}
				strcat( out, buf );
				strcat( out, " " );
				break;
			case VAR_UNDEF:
				if( (c=lookup( sql_buf )) == NULL ) {
					sprintf(error_message,
						"no such field <%s>",
						sql_buf );
					return(0);
					}
				strcat( out, c );
				strcat( out, " " );
				break;
			case STRING:
				sprintf(error_message,"strings not allowed\n");
				return(0);
				break;
			case EQ:
				strcat(out, "=" );
				strcat( out, " ");
				comparison_found=TRUE;
				break;
			case GE:
				strcat(out, ">=" );
				strcat( out, " ");
				comparison_found=TRUE;
				break;
			case GT:
				strcat(out, ">" );
				strcat( out, " ");
				comparison_found=TRUE;
				break;
			case LE:
				strcat(out, "<=");
				strcat(out," ");
				comparison_found=TRUE;
				break;
			case LT:
				strcat(out, "<");
				strcat(out," ");
				comparison_found=TRUE;
				break;
			case NE:
				strcat(out, "ne");
				strcat(out, " ");
				comparison_found=TRUE;
				break;
			case AND:
				strcat(out, "and" );
				strcat(out, " " );
				break;
			case OR:
				strcat(out, "or" );
				strcat(out, " " );
				break;
			case '(':
				strcat(out, LEFT_PAREN );
				strcat(out, " " );
				break;
			case ')':
				strcat(out, RIGHT_PAREN );
				strcat(out, " " );
				break;
			case '+':
				strcat(out, "+" );
				strcat(out, " " );
				break;
			case '-':
				strcat(out, "-" );
				/* note no trailing space */
				break;
			case '*':
				strcat(out, "*" );
				strcat(out, " " );
				break;
			case '/':
				strcat(out, "/" );
				strcat(out, " " );
				break;
			case '!':
				sprintf(error_message,"no not operator");
				return(0);
				break;
			case SQL_ERROR:
				sprintf(error_message,"%s",sql_error_message);
				return(0);
				break;
			default:
				sprintf(error_message,"unknown operator\n");
				return(0);
				break;
			}
		if( strlen(out)-len > 60 ) {
			len=strlen(out);
			strcat(out,"\n");
			}
		}

	if( comparison_found ) {
		error_message[0]='\0';
		return(1);
		}
	else {
		strcpy(error_message,"no comparison in expression");
		return(0);
		}
	}

int sql_lex() /* from yylex, needs no changes */
{
	int c;

	jmp_buf sql_env;
	if( setjmp(sql_env) ) return(SQL_ERROR);


	/* skip leading white spaces */
	while ((c= sql_getc()) == ' ' || c == '\t');

	/* end of string */
	if (c == '\0')
		return(EOS);

	/* number */
	if ((c == '.') || (isdigit(c))) {
		sql_ungetc();
		sql_scanf(&sql_number);
		return(NUMBER);
	}

	/* built-in function, constant, or user variable */
	if (isalpha(c)) {
		char sbuf[100], *p= sbuf;

		do {
			*p++= c;
		} while ((c= sql_getc()) != '\0' && (isalnum(c) || (c == '_')));
		sql_ungetc();
		*p= '\0';
		strcpy(sql_buf, sbuf);
		return(VAR_UNDEF);
	}

	/* string literal */
	if (c == '"') {
		char sbuf[100], *p;

		for (p= sbuf; (c= sql_getc()) != '"'; p++) {
			if (c == '\0')
				sql_error("missing quote");
			if (p >= sbuf+sizeof(sbuf)-1) {
				*p= '\0';
				sql_error("string too long");
			}
			*p= sql_backslash(c);
		}
		*p= 0;
		strcpy(sql_buf, sbuf);
		return(STRING);
	}

	/* operators */
	switch (c) {
	case '=':
		return(sql_follow('=', EQ, '='));
		break;
	case '>':
		return(sql_follow('=', GE, GT));
		break;
	case '<':
		return(sql_follow('=', LE, LT));
		break;
	case '&':
		return(sql_follow('&', AND, '&'));
		break;
	case '|':
		return(sql_follow('|', OR, '|'));
		break;
	case '!':
		return(sql_follow('=', NE, NOT));
		break;
	default:
		return(c);
		break;
	}
}

sql_follow(expect, ifyes, ifno)
int expect, ifyes, ifno;
{
	int c= sql_getc();

	if (c == expect)
		return(ifyes);
	else {
		sql_ungetc();
		return(ifno);
	}
}

sql_backslash(c)
/* interprets "...\n", etc. literals */
int c;
{
	char *index();
	static char transtab[]= "b\bf\fn\nr\rt\t";

	if (c != '\\')
		return(c);
	c= sql_getc();
	if (islower(c) && index(transtab, c))
		return(index(transtab, c)[1]);
	return(c);
}

sql_error(s)
/* called by yyparse(), presumably due to a syntax error */
char *s;
{
	strcpy(sql_error_message,s);
	longjmp(sql_env, 1);
}

int
sql_getc()
{
        return((int) *sql_str_ptr++);
}
 
sql_ungetc()
{
        sql_str_ptr--;
}

sql_scanf(d)
double *d;
{
        int c;
 
        (void) sscanf(sql_str_ptr, "%lf", d);
        while (isdigit(c= sql_getc()));
        if (c == '.')
                while (isdigit(c= sql_getc()));
        if ((c != 'e') && (c != 'E')) {
                sql_ungetc();
                return;
        }
        c= sql_getc();
        if ((c == '+') || (c == '-')) {
                c= sql_getc();
                if (!isdigit(c))
                        sql_error("sql_scanf error");
                else {
                        while (isdigit(c= sql_getc()));
                        sql_ungetc();
                        return;
                }
        }
        else if (isdigit(c)) {
                while (isdigit(c= sql_getc()));
                sql_ungetc();
                return;
        }
        else
                sql_error("sql_scanf error");
}

char *constant_lookup(p)
char *p;
{
	/* parser numerical constant -> number */
	int i;
	for( i=0; i<n_constants; i++ )
		if( !strcmp(p,constant_symbol[i]) )
			return(constant_value[i]);
	return( NULL );
	}


extern char *field_lookup();

char *lookup(p)
char *p;
{
	/* givewn a symbol, returns another symbol:
		case a: view_field_name -> table.field
		case b: parser numerical constant -> number */

	char *constant_lookup(), *c;

	if( (c=constant_lookup(p)) != NULL ) {
		return(c);
		}
	else {
		return( field_lookup(p) );
		}
	}


